# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Definition of constructors that can be added to the python yaml package
"""

from __future__ import absolute_import, annotations, division, print_function

import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml
from yaml import FullLoader, Loader, Node, UnsafeLoader

from config_builder.utils import prepare_config_path

logger = logging.getLogger(__name__)


def join_string(loader: Union[Loader, FullLoader, UnsafeLoader], node: Node) -> Any:
    """
    Define a custom tag handler that can be added as constructor to
    the python yaml package.

    Concatenates all the strings that are given by the node parameter.

    Args:
        loader: a yaml loader needed to parse the content of the given node into
                a sequence
        node: The node for which to apply this constructor

    Returns:
        The concatenated string
    """

    seq = loader.construct_sequence(node)  # type: ignore[arg-type]

    logger.debug(f"join string for a configuration item: {seq}")

    return "".join([str(i) for i in seq])


def join_string_with_delimiter(
    loader: Union[Loader, FullLoader, UnsafeLoader], node: Node
) -> Any:
    """
    Define a custom tag handler that can be added as constructor to
    the python yaml package.

    Concatenates all the strings that are given by the node parameter. In between
    these strings a delimiter is put. The delimiter is defined by the first
    element of the parsed sequence.

    Args:
        loader: a yaml loader needed to parse the content of the given node into
                a sequence
        node:  The node for which to apply this constructors

    Returns:
        The concatenated string
    """
    seq = loader.construct_sequence(node)  # type: ignore[arg-type]

    delimiter = str(seq[0])

    logger.debug(f"join string with delimiter {delimiter} a configuration item: {seq}")

    joined_string = ""
    for index in range(1, len(seq) - 1):
        joined_string += str(seq[index]) + delimiter

    joined_string += str(seq[-1])

    return joined_string


def join_path(loader: Union[Loader, FullLoader, UnsafeLoader], node: Node) -> Any:
    """
    Define a custom tag handler that can be added as constructor to
    the python yaml package.

    Constructs an path out of all given strings by the given node.

    Args:
        loader: a yaml loader needed to parse the content of the given node into
                a sequence
        node: The node for which to apply this constructors

    Returns:
        The constructed path
    """

    seq = loader.construct_sequence(node)  # type: ignore[arg-type]

    logger.debug(f"join path for a configuration item: {seq}")

    joined_path = ""
    for path in seq:
        joined_path = os.path.join(joined_path, str(path))

    return str(Path(joined_path))


def join_object(loader: Union[Loader, FullLoader, UnsafeLoader], node: Node) -> Any:
    """
    Define a custom tag handler that can be added as constructor to
    the python yaml package.

    Allows to include the content of another yaml file.
    The node has to contain two parameters:
    - 1.: The first defines the name of the attribute whose content should be
      extracted from the given yaml file. When the key is an empty string, the whole
      content of the yaml file will be used
    - 2.: The path to the yaml file from which the content should be extracted

    Args:
        loader: a yaml loader needed to parse the content of the given node into
                a sequence
        node: The node for which to apply this constructors
    Returns:
        A dictionary containing the content of the given yaml file
    """

    seq = loader.construct_sequence(node)  # type: ignore[arg-type]

    if len(seq) < 2:
        raise ValueError(
            "You have two provide two parameters when using !join_object:\n"
            " -1.: The first defines the name of the attribute whose content should be "
            " extracted from the given yaml file\n"
            " -2.: The path to the yaml file from which the content should be extracted\n"
        )

    class_type = str(seq[0])

    transformed_data: Union[List[Any], Dict[Any, Any]]

    transformed_data_list: List[Union[List[Any], Dict[Any, Any]]] = []

    for index in range(1, len(seq)):
        config_path = str(Path(str(seq[index])))
        data, config_path = __get_config_data(config_path=config_path)

        if data is not None and class_type != "":
            logger.debug(
                f"Parsed configuration via !join_object "
                f"for class-type '{class_type}' from '{config_path}'"
            )

            _transformed_data = data[class_type]
        else:
            if data is not None and class_type == "":
                logger.debug(
                    f"Parsed configuration via !join_object "
                    f"with emtpy class-type from '{config_path}'"
                )

                _transformed_data = data
            else:
                _transformed_data = {
                    "FileNotFoundError": f"Could not parse content from the "
                    f"configuration file {config_path} "
                    f"for class '{class_type}'!"
                }

        transformed_data_list.append(_transformed_data)

    if len(transformed_data_list) > 1:
        # In case of multiple filepaths given by !join_object,
        # we assume that the respective attribute is of type list.
        # Therefore, we concatenate all the entries to one joined list.
        transformed_data = []
        for _transformed_data in transformed_data_list:
            transformed_data.extend(_transformed_data)
    elif len(transformed_data_list) == 1:
        # In case of a single item, take the entry as is.
        transformed_data = transformed_data_list[0]
    else:
        # This is just a fallback. With checking 'len(seq) < 2' above,
        # we normally have at least one entry in the transformed_data_list
        transformed_data = {}

    return transformed_data


def join_object_from_config_dir(
    loader: Union[Loader, FullLoader, UnsafeLoader], node: Node
) -> Any:
    """
    Define a custom tag handler that can be added as constructor to
    the python yaml package.

    Allows to include the content of another yaml file, that is searched in different directories.
    The node has to contain three parameters:
    -1.: the name of the attribute whose content should be extracted from the given yaml file
    -2.: The name of the configuration file
    -3. to n: directories where to search for the given configuration file

    Args:
        loader: a yaml loader needed to parse the content of the given node into
                a sequence
        node: The node for which to apply this constructors

    Returns:
         An Dict containing the content of the given yaml file
    """

    sequences = loader.construct_sequence(node)  # type: ignore[arg-type]

    if len(sequences) < 3:
        raise ValueError(
            "For !join_object_from_config_dir at least 3 Arguments have to be provided!\n"
            " -1.: class_type\n"
            " -2.: config_file_name\n"
            " -3. to n: directories where to search for the config_file_name"
        )

    class_type = str(sequences[0])
    config_file_name = str(sequences[1])
    config_dirs = sequences[2:]

    transformed_data: Optional[Dict[Any, Any]] = None

    for _, config_dir in enumerate(config_dirs):
        if transformed_data is not None:
            break

        data, config_path = __get_config_data(
            config_path=str(Path(os.path.join(str(config_dir), config_file_name)))
        )

        if data is not None and class_type != "":
            logger.debug(
                f"Parsed configuration via !join_object_from_config_dir "
                f"for class-type '{class_type}' from '{config_path}'"
            )

            transformed_data = data[class_type]
        else:
            if data is not None and class_type == "":
                logger.debug(
                    f"Parsed configuration via !join_object_from_config_dir "
                    f"with emtpy class-type from '{config_path}'"
                )

                transformed_data = data

    if transformed_data is None:
        transformed_data = {
            "FileNotFoundError": f"Could not find a valid "
            f"configuration file for class '{class_type}' "
            f"while parsing for config_file_name '{config_file_name}' "
            f"and config_dirs '{config_dirs}'!"
        }

    return transformed_data


# register the tag handlers
yaml.add_constructor("!join_string", join_string)
yaml.add_constructor("!join_string_with_delimiter", join_string_with_delimiter)
yaml.add_constructor("!join_path", join_path)
yaml.add_constructor("!join_object", join_object)
yaml.add_constructor("!join_object_from_config_dir", join_object_from_config_dir)


def __parse_from_config_path(config_path: str) -> Optional[Dict[Any, Any]]:
    parsed_dict: Optional[Dict[Any, Any]] = None

    if os.path.isfile(config_path):
        with open(config_path, encoding="utf8") as config_file:
            parsed_dict = yaml.load(config_file, Loader) or {}
    else:
        logger.debug(
            "The given config-path does not exist! Can not parse any data. "
            f"config_path: {config_path}"
        )

    return parsed_dict


def __get_config_data(config_path: str) -> Tuple[Optional[Dict[Any, Any]], str]:
    parsed_dict: Optional[Dict[Any, Any]]

    parsed_dict = __parse_from_config_path(config_path=config_path)

    if parsed_dict is not None:
        return parsed_dict, config_path
    else:
        logger.debug(
            f"Could not parse data from '{config_path}'! "
            f"Trying to build a valid config-path "
            f"by using placeholders as substitutes."
        )

        config_path = prepare_config_path(
            config_path=config_path,
        )

        parsed_dict = __parse_from_config_path(config_path=config_path)

        return parsed_dict, config_path



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Definition and handling of a replacement-map that is used in the ConfigBuilder
to provide information that is used to apply placeholder replacement in strings.
"""

import copy
import logging
import os
from threading import local
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


class ReplacementMapLocal(local):
    def __init__(self) -> None:
        local.__init__(self)
        self.replacement_map: Dict[str, str] = {}


__replacement_map_local = ReplacementMapLocal()


def get_current_replacement_map() -> Dict[str, str]:
    """
    Get a copy of the internally used replacement_map.

    Returns:
        copy of the internally used replacement_map
    """

    return copy.deepcopy(__replacement_map_local.replacement_map)


def get_replacement_map_keys() -> List[str]:
    """
    Get the dictionary keys of the internally used replacement_map.

    Returns:
        the list of dictionary keys of the internally used replacement_map
    """

    return list(__replacement_map_local.replacement_map.keys())


def get_replacement_value(key: str) -> Optional[str]:
    """
    Get the value to the given key of the internally used replacement_map.

    Args:
        key: The key for which the value should be returned

    Returns:
        the value to the given key of the internally used replacement_map
    """

    if key in __replacement_map_local.replacement_map:
        return __replacement_map_local.replacement_map[key]

    return None


def set_replacement_map_value(key: str, value: Optional[str]) -> None:
    """
    Add a key value pair to the internally used replacement_map

    Args:
        key: key to be added to the replacement_map
        value: value to be added to the replacement_map

    Returns:
        None
    """

    if (
        key in __replacement_map_local.replacement_map
        and value != __replacement_map_local.replacement_map[key]
    ) or key not in __replacement_map_local.replacement_map:
        if value is not None:
            logger.info(f"Set placeholder value '{key}' to '{value}'")

            __replacement_map_local.replacement_map[key] = value
        else:
            logger.info(f"Reset placeholder value '{key}' to ''")
            __replacement_map_local.replacement_map[key] = ""


def update_replacement_map_value_from_os(key: str) -> None:
    """
    Updates the value of the given key for the internally used replacement_map.
    For the update the value of the os environment variable is used, that matches
    the given key.

    Args:
        key: The key for which to update the value

    Returns:
        None
    """
    if os.getenv(key=key) is not None:
        set_replacement_map_value(key=key, value=os.getenv(key=key))


def update_replacement_map_from_os() -> None:
    """
    Calls the "update_replacement_map_value_from_os" method for every key of the
    internally used replacement map. This realizes the update of the value for any
    key that is set as os environment variable

    Returns:
        None
    """
    logger.debug("Update placeholder values from OS environment variables")
    for key in __replacement_map_local.replacement_map.keys():
        update_replacement_map_value_from_os(key=key)


def clear_replacement_map() -> None:
    """
    Clears the content of the internally used replacement_map

    Returns:
        None
    """
    __replacement_map_local.replacement_map.clear()


def clear_replacement_map_values() -> None:
    """
    Set every value of the internally used replacement_map to None

    Returns:
        None
    """
    for key in __replacement_map_local.replacement_map.keys():
        __replacement_map_local.replacement_map[key] = ""



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

""" Main class for building configuration classes """

import argparse
import logging
import os
from collections import OrderedDict
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type, cast

import yaml
from related import from_yaml, to_model

from config_builder import BaseConfigClass
from config_builder.replacement_map import (
    clear_replacement_map,
    get_current_replacement_map,
    set_replacement_map_value,
    update_replacement_map_from_os,
)
from config_builder.utils import prepare_config_path
from config_builder.yaml_constructors import (
    join_object,
    join_object_from_config_dir,
    join_path,
    join_string,
    join_string_with_delimiter,
)

logger = logging.getLogger(__name__)


class ConfigBuilder:
    def __init__(
        self,
        class_type: Type[BaseConfigClass],
        yaml_config_path: Optional[str] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
        use_argparse: bool = False,
        configure_argparse: Optional[Callable[[argparse.ArgumentParser], None]] = None,
    ):
        """
        Build a configuration object of the given class type. There are three options:

        - Build a configuration object from the given yaml path
        - Build a configuration object by utilizing the argparse
          parameter "--yaml_config_path" and --replacement-config-path

        After a configuration object is build a post-processing takes place,
        consisting of two parts:

        1. Recursively check the class tree for mutual attributes that are defined via the
           parameter _mutual_attributes function of each BaseConfigClass

        2. Recursively check any string attribute (including strings in lists and dictionaries),
           for placeholders (keys of the string_replacement_map) and replacement them
           (values of the string_replacement_map).

        Args:
            class_type: The type of the configuration object
            yaml_config_path: (Optional) A yaml filepath where to build the configuration
                              object from
            string_replacement_map: (Optional) That defines any string placeholders which can
                                    be used in string attributes
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
            use_argparse: Whether to parse the relevant parameters from commandline
            configure_argparse: (Optional) Reference to a method that allows additional
                                argparse configurations
        """

        self.class_type = class_type
        self.yaml_config_path: Optional[str] = yaml_config_path
        self.string_replacement_map = string_replacement_map
        self.no_checks = no_checks
        self.args: Optional[argparse.Namespace] = None
        self.parser: Optional[argparse.ArgumentParser] = None

        # The build of a configuration object should not be affected from any side effect,
        # therefore clear the global replacement map before each run
        clear_replacement_map()

        ConfigBuilder.__update_replacement_map(
            string_replacement_map=self.string_replacement_map
        )

        if use_argparse:
            # Ensure that ConfigBuilder specific argparse parameter are set
            self.parser = self.setup_argparse(
                configure_argparse=configure_argparse,
            )
            self.args = self.parser.parse_args()

            logger.info(f"Parsed arguments from command-line: {self.args}")

            self.yaml_config_path = (
                str(Path(self.args.yaml_config_path))
                if self.args.yaml_config_path
                else None
            )

            # Check any placeholder of the replacement-map whether they have
            # an entry in the parsed arguments
            ConfigBuilder.__update_placeholder_from_argparse(args=self.args)

            # Parse a replacement-map if the filepath is given as commandline argument
            ConfigBuilder.__update_replacement_map(
                string_replacement_map=self.__parse_replacement_map(
                    replacement_config_path=(
                        str(Path(self.args.replacement_config_path))
                        if self.args.replacement_config_path
                        else None
                    )
                )
            )

        if self.yaml_config_path is None:
            raise ValueError(f"Can not build a configuration, yaml-config-path is None")

        # Update the placeholder value in order to be able to prepare the yaml-config-path,
        # in case the path already contains a placeholder value
        update_replacement_map_from_os()
        self.yaml_config_path = prepare_config_path(
            config_path=self.yaml_config_path,
        )

        self.configuration: BaseConfigClass = self.__postprocess_config(
            configuration=ConfigBuilder.__build_from_yaml(
                yaml_config_path=self.yaml_config_path,
                class_type=self.class_type,
            ),
            no_checks=self.no_checks,
        )

    @staticmethod
    def build_configuration_from_dict(
        class_type: Type[BaseConfigClass],
        config_dict: Dict[str, Any],
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> BaseConfigClass:
        """
        Build a configuration object of the given class type from a dictionary.

        Args:
            config_dict: The dictionary to build the configuration object from
            class_type: The type of the configuration object
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
            string_replacement_map: (Optional) That defines any string placeholders which can
                                    be used in string attributes

        Returns:
            The built configuration object
        """
        # The build of a configuration object should not be affected from any side effect,
        # therefore clear the global replacement map before each run
        clear_replacement_map()

        ConfigBuilder.__update_replacement_map(
            string_replacement_map=string_replacement_map
        )

        return ConfigBuilder.__postprocess_config(
            configuration=to_model(class_type, config_dict), no_checks=no_checks
        )

    @staticmethod
    def __parse_replacement_map(
        replacement_config_path: Optional[str],
    ) -> Optional[Dict[str, str]]:
        if replacement_config_path is None:
            logger.warning(
                "The replacement_config_path is None, string replacement will not be available!"
            )
            return None

        if not os.path.isfile(replacement_config_path):
            logger.warning(
                "The given replacement_config_path does not exist: %s, "
                "string replacement will not be available!" % replacement_config_path
            )
            return None

        logger.info("Load replacement-map from: %s" % replacement_config_path)
        with open(
            file=replacement_config_path, mode="r", encoding="utf8"
        ) as config_file:
            return cast(
                Dict[str, str], yaml.load(stream=config_file, Loader=yaml.Loader)
            )

    @staticmethod
    def __update_replacement_map(
        string_replacement_map: Optional[Dict[str, str]],
    ) -> None:
        """
        Updates the global replacements map by calling set_replacement_map_value(...)
        for every entry of the given string-replacement-map

        Args:
            string_replacement_map: The replacement map that should be used to update the
                                    global replacement map

        Returns:
            None
        """

        if string_replacement_map is not None:
            logger.debug("Update placeholder values from given string replacement map")
            for key, value in string_replacement_map.items():
                set_replacement_map_value(key=key, value=value)

    @staticmethod
    def __build_from_yaml(
        yaml_config_path: str,
        class_type: Type[BaseConfigClass],
    ) -> BaseConfigClass:
        """
        Build a configuration object of the given type from the given yaml filepath.

        Args:
            yaml_config_path: The yaml filepath where to build the configuration object from
            class_type: The type of the configuration object

        Returns:
            The build configuration object
        """

        logger.info(
            f"Build config for class type '{class_type}' "
            f"from config path '{yaml_config_path}'"
        )

        if not os.path.isfile(yaml_config_path):
            raise ValueError(
                f"Config file path does not exist. The given path is: {yaml_config_path}"
            )

        with open(yaml_config_path, encoding="utf8") as yaml_file:
            original_yaml = yaml_file.read().strip()

        # register the tag handler
        yaml.add_constructor("!join_string", join_string)
        yaml.add_constructor("!join_string_with_delimiter", join_string_with_delimiter)
        yaml.add_constructor("!join_path", join_path)
        yaml.add_constructor("!join_object", join_object)
        yaml.add_constructor(
            "!join_object_from_config_dir", join_object_from_config_dir
        )

        yml_dict = from_yaml(
            yaml_package=yaml, stream=original_yaml, loader_cls=yaml.Loader
        )

        ConfigBuilder.__recursive_check_dict(yml_dict=yml_dict)

        configuration: BaseConfigClass = to_model(class_type, yml_dict)

        return configuration

    @staticmethod
    def __postprocess_config(
        configuration: BaseConfigClass, no_checks: bool
    ) -> BaseConfigClass:
        """
        Post process a configuration object. The post-processing consists of three steps:

        1. Check attributes to be mutual exclusive (when no_checks is False)
        2. Perform replacements of placeholder variables in the following order:
            1. Use the values of the string-replacement-map from the ConfigBuilder constructor
            2. Update the values from configuration attributes that match a placeholder name
            3. Update the values from equivalent os-environment-variables
        3. Call the check_values method of every attribute that is
           an instance of the BaseConfigClass (when no_checks is False)

        Args:
            configuration: The configuration object to post process
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called

        Returns:
            The post processed configuration object
        """
        # run recursive check and fill the update the globally defined replacement map
        configuration.recursive_check_mutuality_and_update_replacements(
            no_checks=no_checks,
        )

        # Get values of OS environment variables for all values that are now part of
        # the globally defined replacement map
        update_replacement_map_from_os()

        # Replace any placeholder in strings that are part of the build config object
        configuration.recursive_string_replacement()

        # Save the replacement map that has been used to build this config object
        configuration.string_replacement_map.update(get_current_replacement_map())

        if not no_checks:
            configuration.recursive_check_values()

            if not configuration.check_values():
                raise ValueError(
                    f"Check values for configuration class"
                    f"'{configuration.__class__}' failed!"
                )

        return configuration

    @staticmethod
    def setup_argparse(
        configure_argparse: Optional[Callable[[argparse.ArgumentParser], None]] = None
    ) -> argparse.ArgumentParser:
        """
        Init an argparse instance for the config-builder.
        Use the default or provided description for the help message.

        Args:
            configure_argparse:
        Returns:
             The created parser instance
        """

        parser: argparse.ArgumentParser = argparse.ArgumentParser()

        yaml_config_path_option = "yaml_config_path"
        replacement_config_path_option = "replacement-config-path"

        options = [
            item
            for sublist in [a.option_strings for a in parser._actions]
            for item in sublist
        ]

        if yaml_config_path_option not in options:
            parser.add_argument(
                f"{yaml_config_path_option}",
                help="Load all parameters from the given yaml configuration file",
                type=str,
            )

        if replacement_config_path_option not in options:
            parser.add_argument(
                f"--{replacement_config_path_option}",
                help="Path to the string replacement map that defines "
                "placeholders for YAML configuration files",
                type=str,
            )

        for key in get_current_replacement_map().keys():
            parser.add_argument(
                f"--{key}",
                help=f"'{key}': os environment variable, "
                f"which can be used as placeholder",
                type=str,
                default=None,
            )

        if configure_argparse:
            configure_argparse(parser)

        return parser

    @staticmethod
    def __update_placeholder_from_argparse(args: argparse.Namespace) -> None:
        args_dict = vars(args)

        for os_replacement_key in get_current_replacement_map().keys():
            if (
                os_replacement_key in args_dict
                and args_dict[os_replacement_key] is not None
            ):
                set_replacement_map_value(
                    key=os_replacement_key, value=args_dict[os_replacement_key]
                )

    @staticmethod
    def __recursive_check_dict(yml_dict: Dict[Any, Any]) -> None:
        """
        Check the dictionary that has been parsed by a from_yaml(stream=original_yaml),
        whether there have been failures while reading configs
        from file-paths. This indicates either a real wrong configured
        path or that some OS replacements haven't been configured.

        Args:
            yml_dict: Dict parsed from via "from_yaml(stream=original_yaml)"

        Returns:
             None
        """

        for key, value in yml_dict.items():
            if isinstance(value, OrderedDict) or isinstance(value, dict):
                ConfigBuilder.__recursive_check_dict(yml_dict=value)
            else:
                if "FileNotFoundError" in key:
                    raise FileNotFoundError(value)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Definition of the BaseConfigClass that should be used as superclass for any configuration
class that should be built with the ConfigBuilder.
"""
from __future__ import annotations

import copy
import logging
from typing import Any, Dict, List, Optional, cast

import related
from attr import define, field, fields_dict
from related import TypedSequence

from config_builder.replacement_map import (
    get_current_replacement_map,
    set_replacement_map_value,
)
from config_builder.utils import (
    check_list_type,
    replace_directory_in_path,
    replace_with_os_env,
)

logger = logging.getLogger(__name__)


@define
class BaseConfigClass:
    """
    Superclass for any configuration class that should be built with the ConfigBuilder.
    """

    # During the build, the config-builder searches in the class-tree
    # for attributes which math a key of the given string_replacement_map.
    # When an attribute is found which matches the name of a key, the value of this attribute
    # is put under the according key in the string_replacement_map.
    # The key-value pairs can be used in the configuration files to define string PLACEHOLDER,
    # which are replaced at runtime.
    #
    # string_replacement_map FORMAT:
    #
    # KEY = name of the Attribute
    # VALUE = "" => has to be an empty. It will be filled at runtime
    #
    # Example:
    #
    # string_replacement_map = {
    #   ATTRIBUTE_1: ""
    # }
    #
    #
    __string_replacement_map__: Dict[str, str] = field(init=False)

    def __attrs_post_init__(self) -> None:
        self.__string_replacement_map__ = {}

    @property
    def string_replacement_map(self) -> Dict[str, str]:
        return self.__string_replacement_map__

    def to_dict(self, **kwargs) -> Dict[str, Any]:  # type: ignore[no-untyped-def]
        """
        Convenient method to call the to_dict function of the
        related package. It builds a dictionary out of this
        configuration object

        Args:
            **kwargs:

        Returns:
            The created dict.
        """
        kwargs["suppress_private_attr"] = True
        return related.to_dict(self, **kwargs)  # type: ignore[no-any-return]

    def to_yaml(  # type: ignore[no-untyped-def]
        self, yaml_package, dumper_cls, stream=None, default_flow_style=False, **kwargs
    ) -> None:
        """
        Write this configuration object to a file in yaml format

        Args:
            yaml_package: The yaml package to use for the writing
            dumper_cls: The yaml dumper class to use for the writing
            stream: A stream to the file
            default_flow_style:
            **kwargs

        Returns:
            None
        """
        kwargs["suppress_private_attr"] = True
        related.to_yaml(
            self, yaml_package, dumper_cls, stream, default_flow_style, **kwargs
        )

    def to_json(  # type: ignore[no-untyped-def]
        self, indent=4, sort_keys=True, **kwargs
    ) -> str:
        """
        Creates a string in json format out of this configuration object

        Args:
            indent: The indent to use for the json format
            sort_keys: Whether to sort the keys
            **kwargs:

        Returns:
            The created string
        """
        kwargs["suppress_private_attr"] = True
        return related.to_json(self, indent, sort_keys, **kwargs)  # type: ignore[no-any-return]

    @property
    def _mutual_attributes(self) -> List[str]:
        """
        The mutual-attribute-map defines attribute names of this class that
        are mutual exclusive. These attributes should not be set together for one
        config object.

        Returns:
            The mutual attributes of this configuration object
        """
        return []

    def recursive_string_replacement(
        self, string_replacement_map: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Recursively scans the class-tree for string attributes (including attributes in lists
        and dictionaries) and replaces any occurrence of the keys from the global replacement map
        by the according values.

        Args:
            string_replacement_map: (Optional) An extra string_replacement_map that should be used
                                    to update the global replacement map

        Returns:
            None
        """

        if string_replacement_map is not None:
            for key, value in string_replacement_map.items():
                set_replacement_map_value(key=key, value=value)

        for class_name in fields_dict(self.__class__).keys():
            class_attribute = getattr(self, class_name)
            if class_attribute is not None:
                if isinstance(class_attribute, str):
                    self._replace_string_in_attribute(
                        class_name=class_name,
                    )

                # Check if object is of type List[str]
                elif isinstance(class_attribute, TypedSequence) and check_list_type(
                    obj=cast(List[str], class_attribute), obj_type=str
                ):
                    self._replace_strings_in_attribute_list(
                        class_name=class_name,
                        class_attribute=cast(List[str], class_attribute),
                    )

                elif isinstance(class_attribute, dict):
                    self._replace_strings_in_attribute_dict(
                        class_name=class_name,
                        class_attribute=class_attribute,
                    )

                elif isinstance(class_attribute, TypedSequence):
                    for seq_item in class_attribute:
                        if isinstance(seq_item, BaseConfigClass):
                            seq_item.recursive_string_replacement()

                elif isinstance(class_attribute, BaseConfigClass):
                    class_attribute.recursive_string_replacement()

    def _replace_string_in_attribute(
        self,
        class_name: str,
    ) -> None:
        """
        Check if any key of the global replacement map is contained in the class-name or
        class-attribute. If so, update it accordingly.

        Args:
            class_name: Name of the config attribute

        Returns:
            None
        """

        setattr(self, class_name, replace_with_os_env(getattr(self, class_name)))

        for key, value in get_current_replacement_map().items():
            # Don't run replacements with empty string
            if value == "":
                continue

            # Update the corresponding class_name when it is part of
            # the string_replacement_map
            class_attribute_value = getattr(self, class_name)
            if key == class_name and class_attribute_value != value:
                logger.debug(
                    f"Set attribute value of '{class_name}' "
                    f"from '{class_attribute_value}' "
                    f"to '{value}'"
                )
                setattr(self, class_name, value)

            # Replace the occurrence of the value in the class attribute, if it is
            # a directory
            if key in class_attribute_value:
                logger.debug(
                    f"Replace '{key}' "
                    f"in value '{class_attribute_value}' "
                    f"of attribute {class_name} "
                    f"with '{value}'"
                )
                setattr(
                    self,
                    class_name,
                    replace_directory_in_path(
                        file_path=getattr(self, class_name),
                        replacement_key=key,
                        replacement_value=value,
                    ),
                )

    def _replace_strings_in_attribute_list(
        self,
        class_name: str,
        class_attribute: List[str],
    ) -> None:
        """
        Same as the method '_replace_string_in_attribute',
        but runs the replacement check on a list of string-attributes

        Args:
            class_name: Name of this list attribute
            class_attribute: Value of this list attribute

        Returns:
            None
        """

        new_list: List[str] = copy.deepcopy(class_attribute)

        for index, class_attribute_element in enumerate(class_attribute):

            new_list[index] = replace_with_os_env(class_attribute_element)

            for key, value in get_current_replacement_map().items():
                # Don't run replacements with empty string
                if value == "":
                    continue
                # Replace the occurrence of the value in the class attribute, if it is
                # a directory
                if key in class_attribute_element:
                    logger.debug(
                        f"Replace '{key}' "
                        f"in list element '{class_attribute_element}' "
                        f"of attribute {class_name} "
                        f"with '{value}'"
                    )

                    new_list[index] = replace_directory_in_path(
                        file_path=class_attribute_element,
                        replacement_key=key,
                        replacement_value=value,
                    )

        setattr(self, class_name, new_list)

    def _replace_strings_in_attribute_dict(
        self,
        class_name: str,
        class_attribute: Dict[str, Any],
    ) -> None:
        new_dict = self.__replace_strings_in_attribute_dict_step(
            class_name, class_attribute
        )
        setattr(self, class_name, new_dict)

    @staticmethod
    def __replace_strings_in_attribute_dict_step(
        class_name: str,
        class_attribute: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Same as the method '_replace_string_in_attribute', but runs the replacement
        check on a list of string-attributes

        Args:
            class_name: Name of this dictionary attribute
            class_attribute: Value of this dictionary attribute

        Returns:
            None
        """

        new_dict: Dict[str, Any] = copy.deepcopy(class_attribute)

        for index, (class_attribute_key, class_attribute_value) in enumerate(
            class_attribute.items()
        ):
            if isinstance(new_dict[class_attribute_key], str):

                new_dict[class_attribute_key] = replace_with_os_env(
                    new_dict[class_attribute_key]
                )

                for key, value in get_current_replacement_map().items():
                    # Don't run replacements with empty string
                    if value == "":
                        continue

                    # Replace the occurrence of the value in the class attribute, if it is
                    # a directory
                    if key in new_dict[class_attribute_key]:
                        logger.debug(
                            f"Replace '{key}' "
                            f"in dict entry '{new_dict[class_attribute_key]}' "
                            f"of attribute {class_name} "
                            f"with '{value}'"
                        )
                        new_dict[class_attribute_key] = replace_directory_in_path(
                            file_path=new_dict[class_attribute_key],
                            replacement_key=key,
                            replacement_value=value,
                        )

            elif isinstance(new_dict[class_attribute_key], list):
                for list_index, list_element in enumerate(
                    new_dict[class_attribute_key]
                ):
                    if isinstance(list_element, str):
                        new_list_element = replace_with_os_env(list_element)

                        for key, value in get_current_replacement_map().items():
                            # Don't run replacements with empty string
                            if value == "":
                                continue

                            # Replace the occurrence of the value in the class attribute, if it is
                            # a directory
                            if key in list_element:
                                logger.debug(
                                    f"Replace '{key}' "
                                    f"in dict list entry {list_element} "
                                    f"of {class_name} "
                                    f"with '{value}'"
                                )
                                new_list_element = replace_directory_in_path(
                                    file_path=list_element,
                                    replacement_key=key,
                                    replacement_value=value,
                                )
                        new_dict[class_attribute_key][list_index] = new_list_element
                    elif isinstance(list_element, dict):
                        new_dict[class_attribute_key][list_index] = (
                            BaseConfigClass.__replace_strings_in_attribute_dict_step(
                                class_name=class_attribute_key,
                                class_attribute=list_element,
                            )
                        )
            elif isinstance(new_dict[class_attribute_key], dict):
                new_dict[class_attribute_key] = (
                    BaseConfigClass.__replace_strings_in_attribute_dict_step(
                        class_name=class_attribute_key,
                        class_attribute=class_attribute_value,
                    )
                )

        return new_dict

    def check_values(self) -> bool:
        """
        Placeholder method that is intended to be used to check any attribute
        for correct values.

        Returns:
            Whether the attributes contain valid values or not
        """

        return True

    def check_mutual_exclusive(
        self,
    ) -> None:
        """
        Ensures that the mutual-exclusivity defined by the '_mutual_attributes' of each
        child-attribute of type BaseConfigClass is fulfilled for this configuration object.

        Returns:
            None
        """

        if len(self._mutual_attributes) > 0:
            found_exclusive_items = []

            for mutual_attribute in self._mutual_attributes:
                class_attribute_value = getattr(self, mutual_attribute)
                if mutual_attribute in self._mutual_attributes:
                    if class_attribute_value is not None:
                        found_exclusive_items.append(mutual_attribute)

            if len(found_exclusive_items) > 1:
                raise ValueError(
                    f"mutual exclusive validated for class '{self.__class__.__name__}', "
                    f"found attributes: {found_exclusive_items}"
                )

    def _update_string_replacement(
        self,
    ) -> None:
        """
        Recursively checks if this configuration object contains attributes which match
        the name of the keys  of the given globally defined replacement map
        and sets the appropriate value if it is fulfilled.

        Returns:
            The updated string_replacement_map
        """

        replacement_map = get_current_replacement_map()

        for class_attribute in fields_dict(self.__class__).keys():
            value = getattr(self, class_attribute)

            if (
                class_attribute in replacement_map
                and replacement_map[class_attribute] != value
            ):
                logger.debug(
                    f"Set placeholder '{class_attribute}' "
                    f"from '{replacement_map[class_attribute]}' to "
                    f"'{value}' taken from class type '{self.__class__.__name__}'"
                )
                set_replacement_map_value(key=class_attribute, value=value)

    def recursive_check_mutuality_and_update_replacements(
        self,
        no_checks: bool,
    ) -> None:
        """
        Recursively checks if the configuration object violates against any definition of the
        mutual exclusiveness for its attributes. Besides, it is checked if any attribute name
        it contained in the globally defined string replacement map. If that's the case, the
        value of the replacement map is updated accordingly.

        Args:
            no_checks: Whether to skip the mutual exclusiveness check

        Returns:
            None
        """

        # Update the globally defined string-replacement-map
        self._update_string_replacement()

        if (
            not no_checks
        ):  # Check if config items are mutually exclusive as stated in the definition
            self.check_mutual_exclusive()

        for class_name in fields_dict(self.__class__).keys():
            class_attribute = getattr(self, class_name)
            if class_attribute is not None:
                if isinstance(class_attribute, TypedSequence):
                    for seq_item in class_attribute:
                        if isinstance(seq_item, BaseConfigClass):
                            seq_item.recursive_check_mutuality_and_update_replacements(
                                no_checks=no_checks,
                            )
                else:
                    if isinstance(class_attribute, BaseConfigClass):
                        class_attribute.recursive_check_mutuality_and_update_replacements(
                            no_checks=no_checks,
                        )

    def recursive_check_values(
        self,
    ) -> None:
        """
        Recursively calls the 'check_values(...)' method of this configuration object.

        Raises:
            A ValueError is raised when any call fails

        Returns:
            None
        """

        for class_name in fields_dict(self.__class__).keys():
            class_attribute = getattr(self, class_name)

            if class_attribute is not None:
                if isinstance(class_attribute, TypedSequence):
                    for seq_item in class_attribute:
                        if isinstance(seq_item, BaseConfigClass):
                            seq_item.recursive_check_values()
                            if not seq_item.check_values():
                                raise ValueError(
                                    f"Check for attribute '{class_name}' failed!"
                                )
                else:
                    if isinstance(class_attribute, BaseConfigClass):
                        class_attribute.recursive_check_values()
                        if not class_attribute.check_values():
                            raise ValueError(
                                f"Check for attribute '{class_name}' failed!"
                            )



import json
from collections import OrderedDict
from typing import Any, Dict, List, Union


class TupleEncoder(json.JSONEncoder):
    """
    Customize the json encoding in order to have a tuple save conversion
    """

    __tuple_key__ = "__is_tuple__"
    __tuple_content__ = "content"

    def encode(self, obj: Any) -> str:
        """Return a JSON string representation of a Python data structure.

        >>> from json.encoder import JSONEncoder
        >>> JSONEncoder().encode({"foo": ["bar", "baz"]})
        '{"foo": ["bar", "baz"]}'

        REMARK: The encoding will only work for top level tupels and not for nested tuples

        Args:
            obj: The object to encode

        Returns:
            The encoded object as json string
        """

        def tuple_save_encoding(value: Any) -> Any:
            if isinstance(value, tuple):
                return {
                    TupleEncoder.__tuple_key__: True,
                    TupleEncoder.__tuple_content__: value,
                }
            elif isinstance(value, list):
                return [tuple_save_encoding(e) for e in value]
            elif isinstance(value, dict) or isinstance(value, OrderedDict):
                return {k: tuple_save_encoding(v) for k, v in value.items()}
            else:
                return value

        return str(json.JSONEncoder.encode(self, tuple_save_encoding(value=obj)))

    @staticmethod
    def tuple_save_loading_hook(obj: Any) -> Any:
        """
        Method that is designed to be used as object_hook in json.loads
        when decoding a json string, which has been encoded via TupleEncoder.encode

        Args:
            obj: The object to decode

        REMARK: The decoding will only work for top level tuples and not for nested tuples

        Returns:
            The correct representation of the object
        """
        if TupleEncoder.__tuple_content__ in obj and TupleEncoder.__tuple_key__ in obj:
            return tuple(obj[TupleEncoder.__tuple_content__])
        return obj

    @staticmethod
    def decode(data: Any) -> Any:
        """Dedicated method to decode data that has been encoded via TupleEncoder.encode.

        Typically, the input is a dictionary which is recursively decoded by the respective
        calls of TupleEncoder.decode.

        Args:
            data: The (encoded) data that should be decoded.

        Returns:
            The decoded data.
        """
        if isinstance(data, dict):
            # Decode and return tuple object
            if (
                TupleEncoder.__tuple_content__ in data
                and TupleEncoder.__tuple_key__ in data
            ):
                return tuple(data[TupleEncoder.__tuple_content__])

            new_data = {}
            for key, value in data.items():
                # Decode and tuple object
                if isinstance(value, dict):
                    if (
                        TupleEncoder.__tuple_content__ in value
                        and TupleEncoder.__tuple_key__ in value
                    ):
                        new_data[key] = tuple(value[TupleEncoder.__tuple_content__])
                        continue

                # Recursively call decoding method for each dictionary value
                new_data[key] = TupleEncoder.decode(value)
            return new_data
        elif isinstance(data, list):
            # Recursively call decoding method for each list element
            return [TupleEncoder.decode(element) for element in data]
        else:
            # Nothing to decode, end of recursion
            return data



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

""" Top-level YAML Config Builder package """

# -*- coding: utf-8 -*-
from .base_config_class import BaseConfigClass
from .config_builder import ConfigBuilder

__all__ = ["BaseConfigClass", "ConfigBuilder"]
__version__ = "8.5.0"
__license__ = "OLFL-1.3"
